using System;
using System.Collections.Generic;
using System.Text;
using System.IO.Ports;

using System.IO;
using System.Net.Sockets;
using System.Threading.Tasks;
using System.Net;
using System.Threading;
using System.Windows.Forms;

namespace ModBusMasterApp
{
    class modbus
    {
        public SerialPort serialPort = null; // new SerialPort();   //public SerialPort serialPort = null;
        public string modbusStatus;
        public string send_message;
        public string rec_message;
        private const char Header = ':';
        public string CrLf = "\r\n";

        #region Constructor / Deconstructor
        public modbus()
        {
        }
        ~modbus()
        {
        }
        #endregion

        #region Open / Close Procedures
        public bool Open(string portName, int baudRate, int databits, Parity parity, StopBits stopBits)
        {
            //Ensure port isn't already opened:
            if (!serialPort.IsOpen)
            {
                //Assign desired settings to the serial port:
                serialPort.PortName = portName;
                serialPort.BaudRate = baudRate;
                serialPort.DataBits = databits;
                serialPort.Parity = parity;
                serialPort.StopBits = stopBits;
                //These timeouts are default and cannot be editted through the class at this point:
                serialPort.ReadTimeout = 1000;
                serialPort.WriteTimeout = 1000;

                try
                {
                    serialPort.Open();
                }
                catch (Exception err)
                {
                    modbusStatus = "Error opening " + portName + ": " + err.Message;
                    return false;
                }
                modbusStatus = portName + " opened successfully";
                return true;
            }
            else
            {
                modbusStatus = portName + " already opened";
                return false;
            }
        }
        public bool Close()
        {
            //Ensure port is opened before attempting to close:
            if (serialPort.IsOpen)
            {
                try
                {
                    serialPort.Close();
                }
                catch (Exception err)
                {
                    modbusStatus = "Error closing " + serialPort.PortName + ": " + err.Message;
                    return false;
                }
                modbusStatus = serialPort.PortName + " closed successfully";
                return true;
            }
            else
            {
                modbusStatus = serialPort.PortName + " is not open";
                return false;
            }
        }
        #endregion

        #region CRC Computation
        private void GetCRC(byte[] message, ref byte[] CRC)
        {
            //Function expects a modbus message of any length as well as a 2 byte CRC array in which to 
            //return the CRC values:

            ushort CRCFull = 0xFFFF;
            byte CRCHigh = 0xFF, CRCLow = 0xFF;
            char CRCLSB;

            for (int i = 0; i < (message.Length) - 2; i++)
            {
                CRCFull = (ushort)(CRCFull ^ message[i]);

                for (int j = 0; j < 8; j++)
                {
                    CRCLSB = (char)(CRCFull & 0x0001);
                    CRCFull = (ushort)((CRCFull >> 1) & 0x7FFF);

                    if (CRCLSB == 1)
                        CRCFull = (ushort)(CRCFull ^ 0xA001);
                }
            }
            CRC[1] = CRCHigh = (byte)((CRCFull >> 8) & 0xFF);
            CRC[0] = CRCLow = (byte)(CRCFull & 0xFF);
        }
        #endregion
        
        #region Build Message
        public void BuildMessage(byte address, byte type, ushort start, ushort registers, ref byte[] message)
        {
            //Array to receive CRC bytes:
            byte[] CRC = new byte[2];
            byte[] message2 = new byte[20];
            message2 = null;
            
            if (MyGlobals.serial_mode == MyGlobals.MODBUS_RTU)
            {
                message[0] = address;
                message[1] = type;
                message[2] = (byte)(start >> 8);
                message[3] = (byte)start;
                message[4] = (byte)(registers >> 8);
                message[5] = (byte)registers;

                GetCRC(message, ref CRC);
                message[message.Length - 2] = CRC[0];
                message[message.Length - 1] = CRC[1];
            }
            else   //Ascii - FC3
            {
            }
            send_message = null;
            for (int i = 0; i < message.Length; i++)
            {
                send_message += message[i] + " ";
            }
            //send_message+= null;
        }
        #endregion
        static public byte[] str2bytes(string byteData)
        {
            System.Text.ASCIIEncoding asencoding = new System.Text.ASCIIEncoding();
            return Encoding.Default.GetBytes(byteData);
        }

        static byte[] HexToBytes(string input)
        {
            
             byte[] result = new byte[input.Length / 2];
            for (int i = 0; i < result.Length; i++)
            {
                result[i] = Convert.ToByte(input.Substring(2 * i, 2), 16);
            }
            return result;
        }

        #region Check Response
        private bool CheckResponse(byte[] response)
        {
            //Perform a basic CRC check:
            if (MyGlobals.serial_mode == MyGlobals.MODBUS_RTU)
            {
                byte[] CRC = new byte[2];

                rec_message = null;
                for (int i = 0; i < 2; i++)
                {
                    rec_message += CRC[i];
                }

                GetCRC(response, ref CRC);
                if (CRC[0] == response[response.Length - 2] && CRC[1] == response[response.Length - 1])
                    return true;
                else
                    return false;
            }
            else
            {
            }
        }
        #endregion

        #region Get Response
        private void GetResponse(ref byte[] response)
        {
            //There is a bug in .Net 2.0 DataReceived Event that prevents people from using this
            //event as an interrupt to handle data (it doesn't fire all of the time).  Therefore
            //we have to use the ReadByte command for a fixed length as it's been shown to be reliable.
            for (int i = 0; i < response.Length; i++)
            {
                response[i] = (byte)(serialPort.ReadByte());
            }
        }
        #endregion

        #region Function 16 - Write Multiple Registers
        public bool SendFc16(byte address, ushort start, ushort registers, short[] values)
        {
            //Ensure port is open:
            if (serialPort.IsOpen)
            {
                //Clear in/out buffers:
                serialPort.DiscardOutBuffer();
                serialPort.DiscardInBuffer();
                if (MyGlobals.serial_mode == MyGlobals.MODBUS_RTU)
                {
                    //Message is 1 addr + 1 fcn + 2 start + 2 reg + 1 count + 2 * reg vals + 2 CRC
                    byte[] message = new byte[9 + 2 * registers];
                    //Function 16 response is fixed at 8 bytes
                    byte[] response = new byte[8];

                    //Add bytecount to message:
                    message[6] = (byte)(registers * 2);

                    //Put write values into message prior to sending:
                    for (int i = 0; i < registers; i++)
                    {
                        message[7 + 2 * i] = (byte)(values[i] >> 8);
                        message[8 + 2 * i] = (byte)(values[i]);
                    }
                    //Build outgoing message:
                    BuildMessage(address, (byte)16, start, registers, ref message);

                    //Send Modbus message to Serial Port:
                    try
                    {
                        serialPort.Write(message, 0, message.Length);
                        GetResponse(ref response);
                    }
                    catch (Exception err)
                    {
                        modbusStatus = "Error in write event: " + err.Message;
                        return false;
                    }
                    //Evaluate message:
                    if (CheckResponse(response))
                    {
                        modbusStatus = "Write successful";
                        return true;
                    }
                    else
                    {
                        modbusStatus = "CRC error";
                        return false;
                    }
                }
                else
                {
                    byte[] message2 = new byte[30];
                    message2 = null;

                    //Message is 1 addr + 1 fcn + 2 start + 2 reg + 1 count + 2 * reg vals + 2 CRC
                    byte[] message = new byte[1+2+2+4+4+2+(4 * registers)+2+2];
                    //Function 16 response is fixed at 8 bytes
                    byte[] response = new byte[1+2+2+4+4+2+2];

                    string frame = string.Format("{0:x2}", address);     //string.Format("{0:x1}", Header);
                    frame += string.Format("{0:x2}", 16);
                    frame += string.Format("{0:x4}", start);
                    frame += string.Format("{0:x4}", registers);
                    frame += string.Format("{0:x2}", (registers * 2));

                    //Put write values into message prior to sending:
                    for (int i = 0; i < registers; i++)
                    {
                        frame += string.Format("{0:x4}", (values[i]));
                    }
                    message2 = HexToBytes(frame);
                    byte lrc = GetLRC(message2);
                    //lrc = 242;       //30,29,31
                    string TotFrame = string.Format("{0:x1}", Header) + frame + lrc.ToString("X2") + CrLf;
                    message = str2bytes(TotFrame);
                    send_message = null;
                    for (int i = 0; i < message.Length; i++)
                    {
                        send_message += message[i] + " ";
                    }
                    //Send Modbus message to Serial Port:
                    try
                    {
                        serialPort.Write(message, 0, message.Length);
                        GetResponse(ref response);
                    }
                    catch (Exception err)
                    {
                        modbusStatus = "Error in write event: " + err.Message;
                        return false;
                    }
                    //Evaluate message:
                    if (CheckResponse(response))
                    {
                        modbusStatus = "Write successful";
                        return true;
                    }
                    else
                    {
                        modbusStatus = "CRC error";
                        return false;
                    }
                }
            }
            else
            {
                modbusStatus = "Serial port not open";
                return false;
            }
        }
        #endregion

        #region Function 3 - Read Registers
        public bool SendFc3(byte address, ushort start, ushort registers, ref short[] values)
        {
            //Ensure port is open:
            if (serialPort.IsOpen)
            {
                //Clear in/out buffers:
                serialPort.DiscardOutBuffer();
                serialPort.DiscardInBuffer();
                //Function 3 request is always 8 bytes:
                if (MyGlobals.serial_mode == MyGlobals.MODBUS_RTU)
                {
                    byte[] message = new byte[1 + 1 + 2 + 2 + 2];
                    byte[] response = new byte[1 + 1 + 1 + (2 * registers) + 2];
                    //Build outgoing modbus message:
                    BuildMessage(address, (byte)3, start, registers, ref message);
                    try
                    {
                        serialPort.Write(message, 0, message.Length);
                        GetResponse(ref response);
                    }
                    catch (Exception err)
                    {
                        modbusStatus = "Error in read event: " + err.Message;
                        return false;
                    }
                    if (CheckResponse(response))
                    {
                        //Return requested register values:
                        for (int i = 0; i < (response.Length - 5) / 2; i++)
                        {
                            values[i] = response[2 * i + 3];
                            values[i] <<= 8;
                            values[i] += response[2 * i + 4];
                        }
                        modbusStatus = "Read successful";
                        return true;
                    }
                    else
                    {
                        modbusStatus = "CRC error";
                        return false;
                    }
                }
                else
                {
                    byte[] message = new byte[1 + 2 + 2 + 4 + 4 + 2 + 2];
                    byte[] response = new byte[1+2+2+2+(4*registers)+2+2];

                    //Build outgoing modbus message:
                    BuildMessage(address, (byte)3, start, registers, ref message);
                    try
                    {
                        serialPort.Write(message, 0, message.Length);
                        GetResponse(ref response);
                    }
                    catch (Exception err)
                    {
                        modbusStatus = "Error in read event: " + err.Message;
                        return false;
                    }
                    //Evaluate message:
                    if (CheckResponse(response))
                    {
                        string hexString="";
                        //Return requested register values:
                        for (int i = 0; i < (response.Length - 9) / 4; i++)
                        {
                            hexString = Convert.ToChar(response[2 * i + 7]).ToString(); //(response[2 * i + 7]);
                            hexString += Convert.ToChar(response[2 * i + 8]).ToString();
                            hexString += Convert.ToChar(response[2 * i + 9]).ToString();
                            hexString += Convert.ToChar(response[2 * i + 10]).ToString();

                            values[i] = short.Parse(hexString, System.Globalization.NumberStyles.HexNumber);
                        }
                        modbusStatus = "Read successful";
                        return true;
                    }
                    else
                    {
                        modbusStatus = "CRC error";
                        return false;
                    }
                }
            }
            else
            {
                modbusStatus = "Serial port not open";
                return false;
            }
        }
        #endregion
    }
}

